/* Automatically generated by refactor.pl.
 *
 *
 * CMDNAME: rollbackaccount
 * CMDLEVEL: QCMD_OPER
 * CMDARGS: 2
 * CMDDESC: Roll back password/email changes on an account.
 * CMDFUNC: csa_dorollbackaccount
 * CMDPROTO: int csa_dorollbackaccount(void *source, int cargc, char **cargv);
 * CMDHELP: Usage: rollbackaccount <username> <duration>
 * CMDHELP: Rolls back an account to a previous point, reverting password and email changes.
 */

#include "../chanserv.h"
#include "../../lib/irc_string.h"
#include "../../dbapi/dbapi.h"

#include <stdio.h>
#include <string.h>

void csdb_dorollbackaccount_real(DBConn *dbconn, void *arg) {
  nick *np=getnickbynumeric((unsigned long)arg);
  reguser *rup;
  unsigned int userID;
  char *oldpass, *newpass, *oldemail, *newemail;
  time_t changetime, authtime;
  DBResult *pgres;

  if(!dbconn)
    return;

  pgres=dbgetresult(dbconn);

  if (!dbquerysuccessful(pgres)) {
    Error("chanserv", ERR_ERROR, "Error loading account rollback data.");
    return;
  }

  if (dbnumfields(pgres) != 7) {
    Error("chanserv", ERR_ERROR, "Account rollback data format error.");
    dbclear(pgres);
    return;
  }

  if (!np) {
    dbclear(pgres);
    return;
  }

  if (!(rup=getreguserfromnick(np)) || !UHasOperPriv(rup)) {
    Error("chanserv", ERR_ERROR, "No reguser pointer or oper privs in rollback account.");
    dbclear(pgres);
    return;
  }

  chanservsendmessage(np, "Attempting to rollback account %s:", rup->username);
  while(dbfetchrow(pgres)) {
    userID=strtoul(dbgetvalue(pgres, 0), NULL, 10);
    changetime=strtoul(dbgetvalue(pgres, 1), NULL, 10);
    authtime=strtoul(dbgetvalue(pgres, 2), NULL, 10);
    oldpass=dbgetvalue(pgres, 3);
    newpass=dbgetvalue(pgres, 4);
    oldemail=dbgetvalue(pgres, 5);
    newemail=dbgetvalue(pgres, 6);
    if (strlen(newpass) > 0) {
      setpassword(rup, oldpass);
      chanservsendmessage(np, "Restoring old password (%s -> %s)", newpass, oldpass);
    }
    else if (strlen(newemail) > 0) {
      /* WARNING: lastemail untouched */
      freesstring(rup->email);
      rup->email=getsstring(oldemail, EMAILLEN);
      rup->lastemailchange=changetime;
      chanservsendmessage(np, "Restoring old email (%s -> %s)", newemail, oldemail);
    }
  }
  csdb_updateuser(rup);
  chanservstdmessage(np, QM_DONE);

  dbclear(pgres);
}

void csdb_rollbackaccounthistory(nick *np, reguser* rup, time_t starttime) {
  q9u_asyncquery(csdb_dorollbackaccount_real, (void *)np->numeric,
    "SELECT userID, changetime, authtime, oldpassword, newpassword, oldemail, newemail from chanserv.accounthistory where "
    "userID=%u and changetime>%lu order by changetime desc limit 10", rup->ID, starttime);
}

int csa_dorollbackaccount(void *source, int cargc, char **cargv) {
  reguser *rup, *trup;
  nick *sender=source;
  time_t starttime=getnettime();
  long duration;
  
  if (!(rup=getreguserfromnick(sender)))
    return CMD_ERROR;
  
  if (cargc < 2) {
    chanservstdmessage(sender, QM_NOTENOUGHPARAMS, "rollbackaccount");
    return CMD_ERROR;
  }
  
  if (!(trup=findreguser(sender, cargv[0])))
    return CMD_ERROR;
  
  if (UHasOperPriv(trup)) {
    chanservstdmessage(sender, QM_NOACCESS, "rollbackaccount", cargv[0]);
    return CMD_ERROR;
  }
  
  if (!(duration=durationtolong(cargv[1]))) {
    chanservsendmessage(sender, "Invalid duration.");
    return CMD_ERROR;
  }
  starttime-=duration;
  
  cs_log(sender,"ROLLBACKACCOUNT %s #%s %s", trup->username, rup->username, cargv[1]);
  
  csdb_rollbackaccounthistory(sender, trup, starttime);
  
  return CMD_OK;
}
